#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ("OperationProcessor",)

import abc
from typing import (
    TYPE_CHECKING,
    Optional,
)

if TYPE_CHECKING:
    from neptune.core.components.operation_storage import OperationStorage
    from neptune.internal.operation import Operation


class OperationProcessor(abc.ABC):
    @abc.abstractmethod
    def enqueue_operation(self, op: "Operation", *, wait: bool) -> None: ...

    @property
    def operation_storage(self) -> "OperationStorage":
        raise NotImplementedError()

    def start(self) -> None:
        pass

    def pause(self) -> None:
        pass

    def resume(self) -> None:
        pass

    def flush(self) -> None:
        pass

    def wait(self) -> None:
        pass

    def stop(self, seconds: Optional[float] = None) -> None:
        pass

    def close(self) -> None:
        pass



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["Table"]

from typing import (
    TYPE_CHECKING,
    Any,
    Generator,
    List,
    Optional,
)

from neptune.api.field_visitor import FieldToValueVisitor
from neptune.api.models import (
    Field,
    FieldType,
    LeaderboardEntry,
)
from neptune.exceptions import MetadataInconsistency
from neptune.integrations.pandas import to_pandas
from neptune.internal.backends.neptune_backend import NeptuneBackend
from neptune.internal.container_type import ContainerType
from neptune.internal.utils.logger import get_logger
from neptune.internal.utils.paths import (
    join_paths,
    parse_path,
)
from neptune.typing import ProgressBarType

if TYPE_CHECKING:
    import pandas


logger = get_logger()


class TableEntry:
    def __init__(
        self,
        backend: NeptuneBackend,
        container_type: ContainerType,
        _id: str,
        attributes: List[Field],
    ):
        self._backend = backend
        self._container_type = container_type
        self._id = _id
        self._fields = attributes
        self._field_to_value_visitor = FieldToValueVisitor()

    def __getitem__(self, path: str) -> "LeaderboardHandler":
        return LeaderboardHandler(table_entry=self, path=path)

    def get_attribute_type(self, path: str) -> FieldType:
        for field in self._fields:
            if field.path == path:
                return field.type

        raise ValueError(f"Could not find {path} field")

    def get_attribute_value(self, path: str) -> Any:
        for field in self._fields:
            if field.path == path:
                return self._field_to_value_visitor.visit(field)
        raise ValueError("Could not find {} attribute".format(path))

    def download_file_attribute(
        self,
        path: str,
        destination: Optional[str],
        progress_bar: Optional[ProgressBarType] = None,
    ) -> None:
        for attr in self._fields:
            if attr.path == path:
                _type = attr.type
                if _type == FieldType.FILE:
                    self._backend.download_file(
                        container_id=self._id,
                        container_type=self._container_type,
                        path=parse_path(path),
                        destination=destination,
                        progress_bar=progress_bar,
                    )
                    return
                raise MetadataInconsistency("Cannot download file from attribute of type {}".format(_type))
        raise ValueError("Could not find {} attribute".format(path))

    def download_file_set_attribute(
        self,
        path: str,
        destination: Optional[str],
        progress_bar: Optional[ProgressBarType] = None,
    ) -> None:
        for attr in self._fields:
            if attr.path == path:
                _type = attr.type
                if _type == FieldType.FILE_SET:
                    self._backend.download_file_set(
                        container_id=self._id,
                        container_type=self._container_type,
                        path=parse_path(path),
                        destination=destination,
                        progress_bar=progress_bar,
                    )
                    return
                raise MetadataInconsistency("Cannot download ZIP archive from attribute of type {}".format(_type))
        raise ValueError("Could not find {} attribute".format(path))


class LeaderboardHandler:
    def __init__(self, table_entry: TableEntry, path: str) -> None:
        self._table_entry = table_entry
        self._path = path

    def __getitem__(self, path: str) -> "LeaderboardHandler":
        return LeaderboardHandler(table_entry=self._table_entry, path=join_paths(self._path, path))

    def get(self) -> Any:
        return self._table_entry.get_attribute_value(path=self._path)

    def download(self, destination: Optional[str]) -> None:
        attr_type = self._table_entry.get_attribute_type(self._path)
        if attr_type == FieldType.FILE:
            return self._table_entry.download_file_attribute(self._path, destination)
        elif attr_type == FieldType.FILE_SET:
            return self._table_entry.download_file_set_attribute(path=self._path, destination=destination)
        raise MetadataInconsistency("Cannot download file from attribute of type {}".format(attr_type))


class Table:
    def __init__(
        self,
        backend: NeptuneBackend,
        container_type: ContainerType,
        entries: Generator[LeaderboardEntry, None, None],
    ) -> None:
        self._backend = backend
        self._entries = entries
        self._container_type = container_type
        self._iterator = iter(entries if entries else ())

    def to_rows(self) -> List[TableEntry]:
        return list(self)

    def __iter__(self) -> "Table":
        return self

    def __next__(self) -> TableEntry:
        entry = next(self._iterator)

        return TableEntry(
            backend=self._backend,
            container_type=self._container_type,
            _id=entry.object_id,
            attributes=entry.fields,
        )

    def to_pandas(self) -> "pandas.DataFrame":
        return to_pandas(self)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ("ReadOnlyOperationProcessor",)

from typing import TYPE_CHECKING

from neptune.internal.operation_processors.operation_processor import OperationProcessor
from neptune.internal.warnings import (
    NeptuneWarning,
    warn_once,
)

if TYPE_CHECKING:
    from neptune.internal.operation import Operation


class ReadOnlyOperationProcessor(OperationProcessor):
    def enqueue_operation(self, op: "Operation", *, wait: bool) -> None:
        warn_once("Client in read-only mode, nothing will be saved to server.", exception=NeptuneWarning)



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Utility functions to support ML metadata logging with neptune.ai."""
__all__ = [
    "stringify_unsupported",
    "stop_synchronization_callback",
    "TqdmProgressBar",
    "NullProgressBar",
]

from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    Mapping,
    MutableMapping,
    Optional,
    Type,
    Union,
)

from neptune.internal.init.parameters import DEFAULT_STOP_TIMEOUT
from neptune.internal.types.stringify_value import StringifyValue
from neptune.internal.utils.logger import get_logger
from neptune.typing import ProgressBarCallback

if TYPE_CHECKING:
    from neptune.objects.neptune_object import NeptuneObject


logger = get_logger()


def stringify_unsupported(value: Any) -> Union[StringifyValue, Mapping]:
    """Helper function that converts unsupported values in a collection or dictionary to strings.

    Args:
        value (Any): A dictionary with values or a collection

    Example:
        >>> import neptune
        >>> run = neptune.init_run()
        >>> complex_dict = {"tuple": ("hi", 1), "metric": 0.87}
        >>> run["complex_dict"] = complex_dict
        >>> # (as of 1.0.0) error - tuple is not a supported type
        ... from neptune.utils import stringify_unsupported
        >>> run["complex_dict"] = stringify_unsupported(complex_dict)
    """
    if isinstance(value, MutableMapping):
        return {str(k): stringify_unsupported(v) for k, v in value.items()}

    return StringifyValue(value=value)


def stop_synchronization_callback(neptune_object: "NeptuneObject") -> None:
    """Default callback function that stops a Neptune object's synchronization with the server.

    Args:
        neptune_object: A Neptune object (Run, Model, ModelVersion, or Project) to be stopped.

    Example:
        >>> import neptune
        >>> from neptune.utils import stop_synchronization_callback
        >>> run = neptune.init_run(
        ...     async_no_progress_callback = stop_synchronization_callback
        ... )

    For more information, see:
        https://docs.neptune.ai/api/utils/stop_synchronization_callback/
    """
    logger.error(
        "Threshold for disrupted synchronization exceeded. Stopping the synchronization using the default callback."
    )
    neptune_object.stop(seconds=DEFAULT_STOP_TIMEOUT)


class NullProgressBar(ProgressBarCallback):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        pass

    def update(self, *, by: int, total: Optional[int] = None) -> None:
        pass


class TqdmProgressBar(ProgressBarCallback):
    def __init__(
        self,
        *args: Any,
        description: Optional[str] = None,
        unit: Optional[str] = None,
        unit_scale: bool = False,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)

        from tqdm.auto import tqdm

        unit = unit if unit else ""

        self._progress_bar = tqdm(desc=description, unit=unit, unit_scale=unit_scale, **kwargs)

    def __enter__(self) -> "TqdmProgressBar":
        self._progress_bar.__enter__()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self._progress_bar.__exit__(exc_type, exc_val, exc_tb)

    def update(self, *, by: int, total: Optional[int] = None) -> None:
        if total:
            self._progress_bar.total = total
        self._progress_bar.update(by)
